package com.urbancode.anthill3.plugin.scm

import com.urbancode.anthill3.*
import com.urbancode.anthill3.plugin.scm.changelog.*;

import java.util.Date;
import java.util.TimeZone;
import java.util.regex.Pattern;
import java.text.SimpleDateFormat


public class SCMChangelog extends SCMStep {

    //**************************************************************************
    // CLASS
    //**************************************************************************

    //**************************************************************************
    // INSTANCE
    //**************************************************************************

    String changelogUserField;

    // changelog boundaries
    Date startDate;
    Date endDate;
    String startRevision;
    String endRevision;

    // filters for changeset
    final ChangeSetFilter changeSetFilter = new ChangeSetFilter();

    // log time elapsed
    TimeLog tl = new TimeLog();

    /**
     *
     * @return xml document of changes
     */
    public def execute() {

        long executeStart = tl.now();

        //
        // Get Change Log
        //

        long runLogCommandStart = tl.now();
        def logOutput = runLogCommand();
        tl.log("SCMChangelog-> execute()-> runLogCommand()-> ", runLogCommandStart);

        //
        // Parse Change Log into changesets
        //

        // the result change set data
        long parseLogStart = tl.now();
        List<ChangeSet> changeSets = parseLogIntoChangeSets(logOutput);
        tl.log("SCMChangelog-> execute()-> parseLogIntoChangeSets()-> ", parseLogStart);

        //
        // Construct AHP XML message and upload to AHP
        //

        if (changeSets) {
            ChangeSetXMLHelper xmlHelper = new ChangeSetXMLHelper()
            xmlHelper.repoType = REPO_TYPE;
            tl.log("SCMChangelog-> execute()-> ", executeStart);
            tl.log("*** ===== END ENTRY ===== ***", tl.now());
            return xmlHelper.toXML(changeSets);
        }
        else {
            println "No changes detected"
            tl.log("SCMChangelog-> execute()-> ", executeStart);
            tl.log("*** ===== END ENTRY ===== ***", tl.now());
            return null
        }
    }

    /**
     * Parse an input of changelog command into a list of AHP ChangeSets
     * @param logOutput output from a {@link #runLogCommand()} method
     * @return List of ahp changesets
     */
    public List<ChangeSet> parseLogIntoChangeSets(String logOutput) {

        List<ChangeSet> changeSets = []
        String repoUrl = calculateRepositoryUrl();

        // strip userinfo from module uri
        def r = new URI(repoUrl);
        repoUrl = new URI(r.scheme, null, r.host, r.port, r.path, r.query, r.fragment).toString();

        // parse output
        def entrySepPattern = '\0|'+UNIT_SEPARATOR;
        for (String entry : logOutput.split("(?:$entrySepPattern)\0")) {

            if (!entry) {
                // skip empty string - usually start or end of log output
                continue;
            }
            String[] chunks = entry.split(entrySepPattern);

            int i = 0;
            ChangeSet changeSet = new ChangeSet();
            changeSet.module = repoUrl;
            changeSet.id = chunks[i++];


            String authorName = chunks[i++];
            String authorEmail = chunks[i++];
            long authorTime = chunks[i++].toLong();
            String committerName = chunks[i++];
            String committerEmail = chunks[i++];
            long committerTime = chunks[i++].toLong();

            if (changelogUserField == 'committerName') {
                changeSet.user = committerName;
                changeSet.date = new Date(committerTime * 1000);
            }
            else if (changelogUserField == 'committerEmail') {
                changeSet.user = committerEmail;
                changeSet.date = new Date(committerTime * 1000);
            }
            else if (changelogUserField == 'authorName') {
                changeSet.user = authorName;
                changeSet.date = new Date(authorTime * 1000);

            }
            else if (changelogUserField == 'authorEmail') {
                changeSet.user = authorEmail;
                changeSet.date = new Date(authorTime * 1000);
            }
            else {
                // unknown !!!
                throw new AssertionError("Unknown changelog user field $changelogUserField");
            }

            changeSet.properties.putAll([
                'authorName':authorName,
                'authorEmail':authorEmail,
                'authorUnixTimeSeconds': String.valueOf(authorTime),
                'committerName':committerName,
                'committerEmail':committerEmail,
                'committerUnixTimeSeconds':String.valueOf(committerTime),
            ]);

            changeSet.message = chunks[i++]?.trim();


            // read file paths
            while (i < chunks.length) {
                def changeSetType = chunks[i++].trim(); // first type entry has a newline char
                def filePath = chunks[i++];

                ChangeSetFileEntry fileEntry = new ChangeSetFileEntry();
                fileEntry.revision = changeSet.id;
                fileEntry.type = changeSetType;
                fileEntry.path = filePath;
                changeSet.fileSet << fileEntry;
            }
            changeSets << changeSet;
        }

        // filter results
        changeSets.retainAll{ ChangeSet changeSet ->
            boolean allowAuthor = changeSet.hasAllowedAuthor(changeSetFilter);
            boolean allowAnyPaths = changeSet.hasAllowedPath(changeSetFilter);
            if (allowAuthor && allowAnyPaths) {
                // retain changeset
                return true;
            }
            else {
                def message = new StringBuilder("Changeset ${changeSet.id} skipped because ")
                if (!allowAuthor) {
                    message << "it has excluded author ${changeSet.user}"
                }
                if (!allowAuthor && !allowAnyPaths) {
                    message << " and "
                }
                if (!allowAnyPaths) {
                    message << "it contains only excluded file paths (${changeSet.fileSet.collect{it.path}})"
                }
                println message
                // exclude changeset
                return false;
            }
        }

        return changeSets;
    }

    protected String runLogCommand() {
        // we want  a trailing separator from the change sets
        // use unit-separator %x1F in format string, null (%x00) breaks output on msysgit
        // %B is preferred as it doesn't reformat the commit messages, but %s%n%b will have to suffice
        def format = ['%H', '%an', '%ae', '%at', '%cn', '%ce', '%ct', '%s%n%b', ''].join(UNIT_SEPARATOR_FMT)

        def logCommand = [scmCommand, '--no-pager', 'log', '-z', '--no-renames', '--name-status', "--format=format:$format"]
        String startPoint = startRevision ?: lookupRevision(startDate);
        String endPoint = endRevision ?: lookupRevision(endDate);

        // QP should use Dates, CL uses revisions
        if (startPoint) {
            def range = "${startPoint}..${endPoint ?: 'HEAD'}"
            logCommand << range
            logCommand << getTargetBranch();
            logCommand << '--'

            println "Checking for changes"
            def logOutput = getCommandText(logCommand);
            return logOutput.replaceFirst('\0$', ''); // trim the dangling null
        }
        else {
            println "No start point specified; skipping changelog"
            return ""
        }
    }

    /**
     * Return the fully qualified targeted branch name for changelog. Useful for QuietPeriod to override with
     * remote branch name, while normal changelog will use local branch
     * @return
     */
    protected String getTargetBranch() {
        return branch;
    }

    protected String formatChangelogDate(Date date) {
        def df = this.getDateArgumentFormatter()
        return date ? df.format(date) : null;
    }
}
