package com.urbancode.anthill3.plugin.scm;

import com.urbancode.anthill3.*

import java.io.File;
import java.text.DateFormat;
import java.util.Date;
import java.util.regex.Matcher;

abstract public class SCMStep {

    //**************************************************************************
    // CLASS
    //**************************************************************************

    static public final String REPO_TYPE = 'git'

    /*
     * Use of null (e.g. \0 OR %x00) breaks git log format string under msysgit,
     * so we will use the unit-separator character as an alternative to null when needed.
     */
    static protected final char UNIT_SEPARATOR = '\037'
    static protected final String UNIT_SEPARATOR_FMT = '%x1F'

    //**************************************************************************
    // INSTANCE
    //**************************************************************************

    def out = System.out;

    String scmCommand;
    File directory;

    String remoteName;
    String remoteBaseUrl;
    String remoteUrl;
    String remoteUser;
    String remotePass;
    boolean httpsNoVerify = false;
    String branch;
    String revision;
    String fetchRef; // for gerrit preflight checkout

    public void setHttpsNoVerify(boolean httpsNoVerify) {
        this.httpsNoVerify = httpsNoVerify;
        if (this.httpsNoVerify) {
            println "Setting GIT_SSL_NO_VERIFY=true"
        }
    }

    /**
     * Initialize the workspace with data from the remote repository {@link #remoteUrl} for {@link #branch} up to {@link #revision}
     */
    protected void initWorkspace() {

        TimeLog tl = new TimeLog();

        // create the repository directory if not present
        if (!directory.canonicalFile.mkdirs() && !directory.exists()) {
            throw new Exception("Could not create working directory ${directory}")
        }
        else if (directory.isFile()) {
            throw new Exception("Specified working directory is a file!! ${directory}")
        }

        final def cmdHelper = getCommandHelperBase();

        // Repository does not exist, initialize a new one
        if (!new File(directory, '.git').isDirectory()) {
            println "Create git workspace"
            def initCommand = [getScmCommand(), '--no-pager', 'init', '.']
            long gitInit = tl.now();
            cmdHelper.runCommand('Git init', initCommand);
            tl.log("SCMStep-> initWorkspace()-> git init-> " + initCommand + "-> ", gitInit);
        }

        //
        // Add remote if absent
        //  TODO check remote fetch-url?
        //
        println "Checking for remote $remoteName"
        def remoteList = getCommandText([getScmCommand(), '--no-pager', 'remote']).readLines()
        def repoUrl = calculateRepositoryUrl();
        // disable command helper's println statement to avoid printing out user credentials that could
        // possibly exist in repoUrl
        cmdHelper.quiet = true;
        // construct a URI and String without user credentials to be used for println in logs
        def repoUriNoCredentials = new URI(repoUrl);
        def repoUriNoCredentialsStr = new URI(repoUriNoCredentials.scheme,
                                              null,
                                              repoUriNoCredentials.host,
                                              repoUriNoCredentials.port,
                                              repoUriNoCredentials.path,
                                              repoUriNoCredentials.query,
                                              repoUriNoCredentials.fragment).toString();
        if (!remoteList.contains(getRemoteName())) {
            println "Creating remote $remoteName"
            def remoteCommand = [getScmCommand(), '--no-pager', 'remote', 'add', getRemoteName(), repoUrl]
            println "command: " + getScmCommand() + " --no-pager remote add " + getRemoteName() + " " + repoUriNoCredentialsStr
            long remoteAdd = tl.now();
            cmdHelper.runCommand('Git Remote Add', remoteCommand);
            tl.log("SCMStep-> initWorkspace()-> git remote add-> " + remoteCommand + "-> ", remoteAdd);
        }
        else {
            println "Updating remote $remoteName"
            def remoteCommand = [getScmCommand(), '--no-pager', 'remote', 'set-url', getRemoteName(), repoUrl]
            println "command: " + getScmCommand() + " --no-pager remote set-url " + getRemoteName() + " " + repoUriNoCredentialsStr
            long remoteSetUrl = tl.now();
            cmdHelper.runCommand('Git Remote Set-Url', remoteCommand);
            tl.log("SCMStep->initWorkspace()-> git remote set-url-> " + remoteCommand + "-> ", remoteSetUrl);
        }
        // set quiet mode to false as we are no longer going to print out repoUrl
        cmdHelper.quiet = false;
        //
        // Fetch revisions from remote
        // WARN: if http prompts for authentication this will hang!
        //
        println "Fetching new revisions from $remoteName"
        def fetchCommand = [getScmCommand(), '--no-pager', 'fetch', '--force', '--prune', getRemoteName()]
        long fetchNewRevs = tl.now();
        cmdHelper.runCommand('Git Remote Fetch Branch', fetchCommand);
        tl.log("SCMStep->initWorkspace()-> git remote fetch new revisions-> " + fetchCommand + "-> ", fetchNewRevs);

        if (fetchRef) {
            def ref = remoteTrackingBranch;
            fetchCommand << "+$fetchRef:refs/remotes/$ref"
            long remoteFetch = tl.now();
            cmdHelper.runCommand('Git Remote Fetch Specific Ref', fetchCommand)
            tl.log("SCMStep->initWorkspace()-> git remote fetch specific ref-> " + fetchCommand + "-> ", remoteFetch);
        }
    }

    //
    // Utility Methods
    //

    /**
     * Calculate the repository's complete URL from the source config url and the remoteBaseUrl
     * @return
     */
    protected String calculateRepositoryUrl() {
        def pathURI = new URI(remoteUrl);
        def result = pathURI;
        if (remoteBaseUrl) {
            def prefixURI = new URI(remoteBaseUrl+"/");
            result = prefixURI.resolve(remoteUrl);
        }

        URI repoUri = result.normalize();

        // Embed username and password credentials on remoteUrl
        if (remoteUser) {
            if (repoUri.getScheme().startsWith('http')) {
                def userinfo = remoteUser;
                if (remotePass) {
                    userinfo += ":$remotePass"
                }
                repoUri = new URI(repoUri.scheme, userinfo, repoUri.host, repoUri.port, repoUri.path, repoUri.query, repoUri.fragment)
            }
            else {
                println "warning: can not use username/password with non-http(s) url"
            }
        }
        return repoUri.toString();
    }

    /**
     * Print a message to the output stream of choice.
     * @param msg
     */
    protected void println(String msg) {
        out.println(msg);
    }

    protected String getCommandText(def command) {
        final def cmdHelper = getCommandHelperBase();
        def outputText = "[error capturing output]";
        cmdHelper.runCommand(null, command){ Process proc ->
            proc.out.close() // close stdin
            // capture stdout, forward stderr
            def errThread = proc.consumeProcessErrorStream(out) // forward stderr
            try {
                outputText = proc.text.trim();
            }
            finally {
                errThread.join();
            }
        };

        // sanitize the raw output slightly and print
        def entrySepPattern = '\0|'+UNIT_SEPARATOR;
        println(outputText.replaceAll(entrySepPattern, '\n'));

        return outputText;
    }

    protected DateFormat getDateArgumentFormatter() {
        DateFormat gitDateFormat = new java.text.SimpleDateFormat("yyyy-MM-dd HH:mm:ss.S Z")
        gitDateFormat.timeZone = TimeZone.getTimeZone("GMT")
        return gitDateFormat;
    }

    protected CommandHelper getCommandHelperBase() {
        def result = new CommandHelper(directory);
        if (httpsNoVerify) {
            result.environment()['GIT_SSL_NO_VERIFY'] = 'true';
        }
        return result;
    }

    // Convert date into revision
    protected String lookupRevision(Date date) {
        TimeLog tl = new TimeLog();
        long lookupRevStart = tl.now();
        if (date) {
            String dateString = getDateArgumentFormatter().format(date);

            def remoteBranch = remoteTrackingBranch;
            def revListCommand = [this.getScmCommand(), '--no-pager', 'rev-list', '--first-parent' , '--max-count=1', "--before=$dateString", remoteBranch];
            def result = getCommandText(revListCommand);
            tl.log("SCMStep->lookupRevision()-> " + revListCommand + "-> ", lookupRevStart);
            return result;
        }
        else {
            tl.log("SCMStep->lookupRevision()-> ", lookupRevStart);
            return null;
        }
    }


    //
    // Default Getters
    //

    /**
     * Get the branch to operate upon. Default to "master"
     * @return
     */
    public String getBranch() {
        return this.branch ?: 'master'
    }

    /**
     * Get the remote name to use.  Default to "origin"
     * @return
     */
    public String getRemoteName() {
        return this.remoteName ?: 'origin';
    }

    public String getRemoteTrackingBranch() {
        def ref = fetchRef ?: this.getBranch();
        def shortRef = ref.replaceFirst('^refs/(?:heads/|tags/|remotes/[^/]+/)?', '')
        return this.getRemoteName() + "/" + shortRef
    }

    /**
     * Get the scm command to use. Default to "git"
     * @return
     */
    public String getScmCommand() {
        return this.scmCommand ?: 'git';
    }
}
